
#ifndef HH_SPLIT_HANDLER
#define HH_SPLIT_HANDLER

#include "flow_handler.hh"

#include <errno.h>

class split_handler : public flow_handler
{
    bool conn1, conn2;
    bool done;
    int which;
    address addr;

    string in1, in2;
    int whichin;

    flow_handler *acc_next1;
    flow_handler *acc_next2;

protected:
    virtual void dump_children(string& out, int indent) {
	downstream[0]->dump(out, indent);
	downstream[1]->dump(out, indent);
    }

public:
    split_handler(init_context& ctxt) : flow_handler(ctxt, false) {
	downstream.push_back(ctxt.plumb(*this, AF_INET, SOCK_STREAM));
	downstream.push_back(ctxt.plumb(*this, AF_INET, SOCK_STREAM));

	acc_next1 = acc_next2 = 0;
	done = conn1 = conn2 = false;
	which = 0;
	whichin = 0;
    }

    split_handler(split_handler& from,
		  flow_handler *_next1, flow_handler *_next2) : flow_handler(from)
    {
	downstream.push_back(_next1);
	downstream.push_back(_next2);

	acc_next1 = acc_next2 = 0;
	done = conn1 = conn2 = false;
	which = 0;
	whichin = 0;
    }

    virtual int connect(address a) {
	addr = a;

	sockaddr_in addr2;
	assert(a.addrlen() == sizeof addr2);
	memcpy(&addr2, a.addr(), a.addrlen());
	addr2.sin_port = htons(ntohs(addr2.sin_port) + 1);

	address a2(&addr2, sizeof addr2);
	ts_debug_2("split_handler(1): %s", a.c_str());
	ts_debug_2("split_handler(2): %s", a2.c_str());

	int ret = downstream[0]->connect(a);
	if (ret != 0) {
	    ts_debug_2("- unable to connect 1: %s", strerror(errno));
	    return -1;
	}
	ret = downstream[1]->connect(a2);
	if (ret != 0) {
	    ts_debug_2("- unable to connect 2: %s", strerror(errno));
	    return -1;
	}

	return 0;
    }

    virtual void connected(flow_handler *from, bool success) {
	if (done) return;

	if (!success) {
	    ts_debug_2("- connect failed");
	    done = true;
	    get_upstream().connected(this, false);
	    return;
	}

	if (from == downstream[0]) {
	    ts_debug_2("- conn1 connected!");
	    conn1 = true;
	} else {
	    ts_debug_2("- conn2 connected!");
	    conn2 = true;
	}

	if (conn1 && conn2) {
	    ts_debug_2("- both are connected!");
	    get_upstream().connected(this, true);
	}
    }

    virtual int bind(address a) {
	addr = a;

	sockaddr_in addr2;
	assert(a.addrlen() == sizeof addr2);
	memcpy(&addr2, a.addr(), a.addrlen());
	addr2.sin_port = htons(ntohs(addr2.sin_port) + 1);

	address a2(&addr2, sizeof addr2);
	ts_debug_2("split_handler(1): %s", a.c_str());
	ts_debug_2("split_handler(2): %s", a2.c_str());

	int ret = downstream[0]->bind(a);
	if (ret != 0) {
	    ts_debug_2("- unable to bind 1: %s", strerror(errno));
	    return -1;
	}
	ret = downstream[1]->bind(a2);
	if (ret != 0) {
	    ts_debug_2("- unable to bind 2: %s", strerror(errno));
	    return -1;
	}

	ts_debug_2("- bound!");
	return 0;
    }

    virtual int listen(int backlog) {
	downstream[0]->listen(backlog);
	downstream[1]->listen(backlog);
	return 0;
    }

    virtual acceptret accept()
    {
	if (!acc_next1 || !acc_next2) return acceptret();

        flow_handler* h = new split_handler(*this, acc_next1, acc_next2);
	address a = acc_next1->getpeername();

	acc_next1 = acc_next2 = 0;

	return acceptret(h, a);
    }

    virtual int shutdown(bool r, bool w) {
	downstream[0]->shutdown(r, w);
	downstream[1]->shutdown(r, w);
	return 0;
    }

    virtual void accept_ready(flow_handler *from)
    {
	address a;
	
	assert(from == downstream[0] || from == downstream[1]);
	acceptret acc = from->accept();
	assert(acc);

	if (from == downstream[0]) {
	    ts_debug_2("- got next1");
	    acc_next1 = acc.h;
	} else {
	    ts_debug_2("- got next2");
	    acc_next2 = acc.h;
	}

	if (acc_next1 && acc_next2) {
	    ts_debug_2("- have em both!");
	    get_upstream().accept_ready(this);
	}
    }

    virtual bool avail(flow_handler *from, data d)
    {
	assert(from == downstream[0] || from == downstream[1]);
	if (from == downstream[0]) in1.append(d); else in2.append(d);

	string aa;

	string::iterator in1p = in1.begin();
	string::iterator in2p = in2.begin();

	while (1) {
	    if (whichin == 0) {
		if (in1p == in1.end()) break;
		aa.append(in1p, in1p+1);
		ts_debug_2("- from 1, got char \"%c\"", *in1p);
		++in1p;
		whichin = 1;
	    }
	    if (whichin == 1) {
		if (in2p == in2.end()) break;
		ts_debug_2("- from 2, got char \"%c\"", *in2p);
		aa.append(in2p, in2p+1);
		++in2p;
		whichin = 0;
	    }
	}

	in1.erase(in1.begin(), in1p);
	in2.erase(in2.begin(), in2p);

	if (aa.length())
	    return get_upstream().avail(this, aa);

	return true;
    }

    virtual int write(data d)
    {
	string d1, d2;

	for (unsigned int i = 0; i < d.length(); ++i) {
	    if (which) d2.append(1, d[i]); else d1.append(1, d[i]);
	    which = !which;
	}

	bool good = true;

	if (d1.length()) good = good && downstream[0]->write(d1);
	if (d2.length()) good = good && downstream[1]->write(d2);

	return good;
    }

    DECLARE_HANDLER;
};

#endif
