
#include "flow_handler.hh"
#include "default_handler.hh"

#include <iostream>
#include <vector>

#include "tesla.h"

using namespace std;

flow_handler::infovec flow_handler::handlers;
flow_handler::infomap *flow_handler::reg_handlers;

void flow_handler::register_handler(const info& inf)
{
    // Static initialization trick
    if (reg_handlers == 0) {
	ts_debug_2("Allocating handlers vector");
	reg_handlers = new infomap;
    }
    ts_debug_2("Registering flow_handler %s (domain=%d, type=%d)",
	       inf.get_name().c_str(), inf.get_domain(), inf.get_type());
    (*reg_handlers)[inf.get_name()] = &inf;
}

void flow_handler::make_handlers()
{
    char *h = strdup(getenv("TS_HANDLERS"));
    char *hh = h;
    char *tok;

    if (reg_handlers == 0) {
	ts_error("No handlers registered");
	return;
    }

    while ((tok = strtok(hh, ":")) != 0) {
	infomap::const_iterator i = reg_handlers->find(tok);
	if (i == reg_handlers->end()) {
	    ts_error("Unknown flow_handler %s", tok);
	} else {
	    ts_debug_2("Pushing flow_handler %s (%s) on stack", tok, i->second->get_clazz().c_str());
	    handlers.push_back(i->second);
	}
	hh = 0;
    }
    free(h);
}

flow_handler::info default_info("default", "default", -1, -1, 0);

flow_handler::flow_handler(const init_context& ctxt, bool plumb) : inf(ctxt.get_info())
{
    upstream = &ctxt.get_upstream();
    domain = ctxt.get_domain();
    type = ctxt.get_type();
    level = ctxt.get_level();
    
    if (plumb) {
	if (!ctxt.is_accepted())
	    downstream.push_back(ctxt.plumb(*this, ctxt.get_domain(), ctxt.get_type()));
    }
}

flow_handler *flow_handler::plumb(int domain, int type, int fd)
{
    if (fd < 0)
	return plumb(*this, domain, type, get_level());

    init_context ctxt(default_info, *this, domain, type, handlers.size());
    return new default_handler(ctxt, fd);    
}

acceptret flow_handler::accept()
{
    assert(downstream.size() == 1);
    assert(inf.get_cf());
    acceptret h = downstream[0]->accept();
    if (!h)
	return acceptret();

    init_context ctxt(inf, *(flow_handler*)0, domain, type, level, true);
    flow_handler *s = inf.create(ctxt);
    s->downstream.push_back(h.h);
    h.h->upstream = s;
    
    return acceptret(s, h.addr);
}

flow_handler *flow_handler::plumb(flow_handler& upstream, int domain, int type, int level)
{
    if (handlers.size() == 0) return 0;

    ts_debug_2("plumb, upstream=%p, domain=%d, type=%d, level=%d",
	       &upstream, domain, type, level);

    for (unsigned int i = level; i < handlers.size(); ++i) {
	ts_debug_2("  - i=%d of %d; d,t=%d,%d", i, handlers.size(), handlers[i]->get_domain(), handlers[i]->get_type());
	if (handlers[i]->get_domain() == domain &&
	    handlers[i]->get_type() == type)
	{
	    ts_debug_2("    - got it!");
	    init_context ctxt(*(handlers[i]), upstream, domain, type, i + 1);
	    return handlers[i]->create(ctxt);
	}
    }

    if (level == 0) return 0;

    init_context ctxt(default_info, upstream, domain, type, handlers.size());
    return new default_handler(ctxt);
}

int flow_handler::connect(address a) {
    assert(downstream.size() == 1);
    return downstream[0]->connect(a);
}
int flow_handler::bind(address a) {
    assert(downstream.size() == 1);
    return downstream[0]->bind(a);
}

acceptret flow_handler::accept();
int flow_handler::close() {
    assert(downstream.size() == 1);
    return downstream[0]->close();
}

void flow_handler::accept_ready(flow_handler *from) {
    assert(upstream && downstream.size() == 1 && from == downstream[0]);
    upstream->accept_ready(this);
}
void flow_handler::connected(flow_handler *from, bool success) {
    assert(upstream && downstream.size() == 1 && from == downstream[0]);
    upstream->connected(this, success);
}
bool flow_handler::avail(flow_handler *from, data d) {
    assert(upstream && downstream.size() == 1 && from == downstream[0]);
    return upstream->avail(from, d);
}
int flow_handler::write(data d) {
    assert(downstream.size() == 1);
    return downstream[0]->write(d);
}
int flow_handler::shutdown(bool r, bool w) {
    assert(downstream.size() == 1);
    return downstream[0]->shutdown(r,w);
}
int flow_handler::listen(int backlog) {
    assert(downstream.size() == 1);
    return downstream[0]->listen(backlog);
}

address flow_handler::getsockname() {
    assert(downstream.size() == 1);
    return downstream[0]->getsockname();
}
address flow_handler::getpeername() {
    assert(downstream.size() == 1);
    return downstream[0]->getpeername();
}

string flow_handler::getsockopt(int level, int optname, int maxlen) {
    assert(downstream.size() == 1);
    return downstream[0]->getsockopt(level, optname, maxlen);
}
int flow_handler::setsockopt(int level, int optname, string value) {
    assert(downstream.size() == 1);
    return downstream[0]->setsockopt(level, optname, value);
}


void flow_handler::may_write(flow_handler *from, bool may) {
    assert(upstream && downstream.size() == 1 && from == downstream[0]);
    upstream->may_write(this, may);
}

void flow_handler::may_avail(bool may) {
    for (vector<flow_handler*>::const_iterator i = downstream.begin(); i != downstream.end(); ++i)
	(*i)->may_avail(may);
}
