/*
 * TESLA
 *
 * Jon Salz <jsalz@mit.edu>
 * Alex C. Snoeren <snoeren@lcs.mit.edu>
 *
 * Copyright (c) 2001-2 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. For more information, see the `COPYING' file in the source
 * distribution.
 *
 * $Id: teslamaster.cc,v 1.10 2002/06/06 04:00:25 jsalz Exp $
 *
 * The master process implementation.
 *
 */

#include <fcntl.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/wait.h>
#include <unistd.h>

#include <iostream>
#include <map>
#include <set>

#include "async.hh"
#include "client.hh"
#include "flow_handler.hh"
#include "tesla.h"
#include "tesla-internal.h"
#include "top_handler.hh"

using namespace std;

class top_handler;

set<top_handler*> close_set;
set<client*> clients;
map<int, top_handler*> conns; /* by conn_id */
map<int, top_handler*> conns_by_fd;

int read_fully(int fd, void *buf, int count)
{
    int orig_count = count;
    while (count) {
        int bytes = read(fd, buf, count);
        if (bytes <= 0)
            return 0;
        assert(bytes <= count);
        buf = static_cast<char*>(buf) + bytes;
        count -= bytes;
    }
    return orig_count;
}

int write_fully(int fd, const void *buf, int count)
{
    int orig_count = count;
    while (count) {
        int written = write(fd, buf, count);
        if (written <= 0)
            return 0;
        assert(written <= count);
        buf = static_cast<const char*>(buf) + written;
        count -= written;
    }
    return orig_count;
}

void send_msg(int fd, master_msg_t *msg) {
    ts_debug_2("Sending message: id=%d, conn_id=%d, type=%s", msg->id, msg->conn_id,
	       MSG_TYPE(msg->type));

    int len = sizeof *msg - sizeof msg->id;
    if (write(fd, &len, sizeof len) != sizeof len)
	ts_fatal("Unable to write length");
    if (write(fd, &msg->id, sizeof *msg - sizeof msg->len) != sizeof *msg - sizeof msg->len)
	ts_fatal("Unable to write message body");
}

void send_msg_header(int fd, master_msg_t *msg, int length) {
    ts_debug_2("Sending message: id=%d, conn_id=%d, type=%s", msg->id, msg->conn_id,
	       MSG_TYPE(msg->type));

    msg->len = MASTER_MSG_HDRLEN + length - sizeof msg->len;
    if (!write_fully(fd, msg, MASTER_MSG_HDRLEN))
	ts_fatal("Unable to write length");
}

static struct cmsghdr	*cmptr = NULL;	/* buffer is malloc'ed first time */
#define	CONTROLLEN	(sizeof(struct cmsghdr) + sizeof(int))

int
send_fd(int clifd, int fd) {
    struct iovec	iov[1];
    struct msghdr	msg;
    char buf[2]; /* send_fd()/recv_fd() 2-byte protocol */

    iov[0].iov_base = buf;
    iov[0].iov_len  = 2;
    msg.msg_iov     = iov;
    msg.msg_iovlen  = 1;
    msg.msg_name    = NULL;
    msg.msg_namelen = 0;
    if (fd<0) {
	msg.msg_control    = NULL;
	msg.msg_controllen = 0;
	buf[1] = -fd;	/* nonzero status means error */
	if (buf[1] == 0) buf[1] = 1;	/* -256, etc. would screw up protocol */
    } else {
	if (cmptr == NULL && (cmptr = (cmsghdr*)malloc(CONTROLLEN)) == NULL) ts_fatal("Bollocks!");
	cmptr->cmsg_level  = SOL_SOCKET;
	cmptr->cmsg_type   = SCM_RIGHTS;
	cmptr->cmsg_len    = CONTROLLEN;
	msg.msg_control    = (caddr_t) cmptr;
	msg.msg_controllen = CONTROLLEN;
	*(int *)CMSG_DATA(cmptr) = fd;	/* the fd to pass */
	buf[1] = 0;		/* zero status means OK */
    }
    buf[0] = 0;			/* null byte flag to recv_fd() */
  
    if (sendmsg(clifd, &msg, 0) != 2) ts_fatal("Bollocks! %s", strerror(errno));
    return(0);
}


int next_conn_id = 1000;

void client::ravail() {
    master_msg_t msg;
    size_t bytes;

    bytes = read(fd, &msg.len, sizeof msg.len);
    if (bytes != 4) {
	ts_debug_2("Client %d has closed", fd);
	clients.erase(this);
	delete this;
	return;
    }
    
    if (!read_fully(fd, &msg.id, msg.len)) {
	ts_error("Got wrong number of bytes on %d", fd);
	clients.erase(this);
	delete this;
	return;
    }

    ts_debug_2("Received message: id=%d, conn_id=%d, pid=%d, type=%s",
	       msg.id, msg.conn_id, msg.pid, MSG_TYPE(msg.type));

    switch (msg.type) {
      case MSG_HELLO:
	{
	  msg.merrno = 0;
	  send_msg(fd, &msg);
	  break;
	}

      case MSG_NAK:
	{
	    ts_debug_2("Client %d died", fd);
	    clients.erase(this);
	    delete this;
	    break;
	}

      case MSG_SOCKET:
	{
	    ts_debug_2(" - domain=%d; type=%d", msg.body.socket.domain, msg.body.socket.type);

	    flow_handler *h = flow_handler::plumb(*(flow_handler*)0, msg.body.socket.domain, msg.body.socket.type);

	    if (h) {
		ts_debug_2("   - got a stack!");

		string out;
		h->dump(out);
		ts_debug_2("Handlers:\n%s", out.c_str());

		int pair[2];
		if (socketpair(PF_UNIX, h->get_info().get_type(), 0, pair) != 0)
		    ts_fatal("Unable to socketpair");
		int conn_id = next_conn_id++;
		
		msg.conn_id = conn_id;
		msg.type = MSG_FH_FOLLOWS;
		msg.merrno = 0;
		send_msg(fd, &msg);

		send_fd(fd, pair[0]);
		close(pair[0]);

		if (h->get_info().get_type() != SOCK_DGRAM) {
		    if (fcntl(pair[1], F_SETFL, O_NONBLOCK))
			ts_fatal("fcntl: %s", strerror(errno));
		}

		top_handler *c = new top_handler(pair[1], h->get_info().get_type(), conn_id, h);
		conns[conn_id] = c;
		conns_by_fd[pair[1]] = c;
	    } else {
		ts_debug_2("   - no stack; declining");
		msg.type = MSG_NAK;
		send_msg(fd, &msg);
	    }
	    break;
	}

      case MSG_CONNECT:
	{
	    address a(msg.body.address.addr, msg.body.address.addrlen);
	    map<int, top_handler*>::const_iterator i = conns.find(msg.conn_id);
	    if (i == conns.end())
		ts_fatal("Reference to nonexistent conn_id");

	    int ret = i->second->get_handler()->connect(a);
	    ts_debug_2("Connect request: %d", ret);
	    msg.type = (ret == 0 ? MSG_ACK : MSG_NAK);
	    msg.merrno = -ret;
	    send_msg(fd, &msg);
	    break;
	}

      case MSG_BIND:
	{
	    address a(msg.body.address.addr, msg.body.address.addrlen);
	    map<int, top_handler*>::const_iterator i = conns.find(msg.conn_id);
	    if (i == conns.end())
		ts_fatal("Reference to nonexistent conn_id");

	    int ret = i->second->get_handler()->bind(a);
	    ts_debug_2("Bind request: %d", ret);
	    msg.type = (ret == 0 ? MSG_ACK : MSG_NAK);
	    msg.merrno = -ret;
	    send_msg(fd, &msg);
	    break;
	}

      case MSG_LISTEN:
	{
	    map<int, top_handler*>::const_iterator i = conns.find(msg.conn_id);
	    if (i == conns.end())
		ts_fatal("Reference to nonexistent conn_id");

	    int ret = i->second->get_handler()->listen(5);
	    ts_debug_2("Listen request: %d", ret);
	    msg.type = ret == 0 ? MSG_ACK : MSG_NAK;
	    msg.merrno = -ret;
	    //	    send_msg(fd, &msg);
	    break;
	}

      case MSG_ACCEPT:
	{
	    map<int, top_handler*>::const_iterator i = conns.find(msg.conn_id);
	    if (i == conns.end())
		ts_fatal("Reference to nonexistent conn_id");

	    acceptret h = i->second->get_handler()->accept();

	    if (h) {
		ts_debug_2("   - got a stack!");

		string out;
		h.h->dump(out);
		ts_debug_2("Handlers:\n%s", out.c_str());

		int pair[2];
		if (socketpair(PF_UNIX, h.h->get_type(), 0, pair) != 0)
		    ts_fatal("Unable to socketpair");
		int conn_id = next_conn_id++;

		msg.conn_id = conn_id;
		msg.merrno = 0;
		msg.type = MSG_ACK;
		msg.body.address.addrlen = h.addr.addrlen();
		memcpy(msg.body.address.addr, h.addr.addr(), h.addr.addrlen());
		send_msg(fd, &msg);

		msg.type = MSG_FH_FOLLOWS;
		send_msg(fd, &msg);

		send_fd(fd, pair[0]);
		close(pair[0]);
		
		// All TESLA sockets are non-blocking
		if(fcntl(pair[1], F_SETFL, O_NONBLOCK))
		  ts_fatal("fcntl: %s", strerror(errno));

		top_handler *c = new top_handler(pair[1], h.h->get_type(), conn_id, h.h);
		conns[conn_id] = c;
		conns_by_fd[pair[1]] = c;
	    } else {
		ts_debug_2("   - accept failed");
		msg.type = MSG_NAK;
		send_msg(fd, &msg);
	    }
	    break;
	}

      case MSG_GETSOCKOPT:
	{
	    ts_debug_2("Got getsockopt");

	    map<int, top_handler*>::const_iterator i = conns.find(msg.conn_id);
	    if (i == conns.end())
		ts_fatal("Reference to nonexistent conn_id");

	    string val = i->second->getsockopt(msg.body.sockopt.level,
					       msg.body.sockopt.optname,
					       msg.body.sockopt.optlen);

	    if (val.length()) {
		msg.type = MSG_ACK;
		msg.merrno = 0;
		send_msg_header(fd, &msg, sizeof msg.body.sockopt + val.length());
		msg.body.sockopt.optlen = val.length();
		if (!write_fully(fd, &msg.body.sockopt, sizeof msg.body.sockopt) ||
		    !write_fully(fd, val.data(), val.length()))
		{
		    ts_fatal("Unable to write sockopt data to %d", fd);
		}
	    } else {
		msg.type = MSG_NAK;
		send_msg(fd, &msg);
	    }

	    break;
	}

      case MSG_SETSOCKOPT:
	{
	    ts_debug_2("Got setsockopt");

	    map<int, top_handler*>::const_iterator i = conns.find(msg.conn_id);
	    if (i == conns.end())
		ts_fatal("Reference to nonexistent conn_id");

	    string opt(msg.body.sockopt.optval, msg.body.sockopt.optlen);
	    int ret = i->second->setsockopt(msg.body.sockopt.level, msg.body.sockopt.optname, opt);

	    msg.type = (ret ? MSG_NAK : MSG_ACK);
	    msg.merrno = -ret;
	    send_msg(fd, &msg);

	    break;
	}

      case MSG_GETSOCKNAME:
      case MSG_GETPEERNAME:
	{
	    ts_debug_2("Got get%sname", msg.type == MSG_GETSOCKNAME ? "sock" : "peer");

	    map<int, top_handler*>::const_iterator i = conns.find(msg.conn_id);
	    if (i == conns.end())
		ts_fatal("Reference to nonexistent conn_id");

	    address val = msg.type == MSG_GETSOCKNAME ? i->second->getsockname() :
		i->second->getpeername();

	    if (val.addrlen()) {
		msg.type = MSG_ACK;
		msg.merrno = 0;
		msg.body.address.addrlen = val.addrlen();
		memcpy(msg.body.address.addr, val.addr(), val.addrlen());

		send_msg(fd, &msg);
	    } else {
		msg.type = MSG_NAK;
		send_msg(fd, &msg);
	    }

	    break;
	}

    default:
      ts_fatal("Unexpected message type");
    }
}

pid_t pid;

#if 0
extern "C"
void pipe_sigaction(int signal, siginfo_t *info, void *arg)
{
    assert(signal == SIGPIPE && info);

    ts_debug_1("Received SIGPIPE on %d", info->si_fd);
    return;

    map<int, top_handler*>::const_iterator i = conns.find(info->si_fd);
    if (i == conns.end())
	return;

    ts_debug_2(" - second SIGPIPE to top_handler");

    // SIGPIPE from writing on a connection to the client
    i->second->sigpipe();
}
#endif

int main(int argc, const char* argv[])
{
    if (argc != 3)
	ts_fatal("Two arguments are required");

    flow_handler::make_handlers();

#if 0
    struct sigaction act;
    memset(&act, 0, sizeof act);
    act.sa_sigaction = pipe_sigaction;
    act.sa_flags = SA_RESTART | SA_SIGINFO;
    sigaction(SIGPIPE, &act, 0);
#else
    signal(SIGPIPE, SIG_IGN);
#endif

    pid = atoi(argv[1]);

    int fd = atoi(argv[2]);
    assert(fd >= 0 && fd < FD_SETSIZE);
    ts_debug_2("Listening on %d", fd);
    clients.insert(new client(fd));

    while (clients.size()) {
	ts_debug_2("Main loop");
	async::select();

	for (set<top_handler*>::const_iterator i = close_set.begin(); i != close_set.end(); ++i) {
	    (*i)->close();
	    delete *i;
	}

	close_set.clear();
    }

    int status;
    wait(&status);
    return WEXITSTATUS(status);
}
